/*
 * impopen.c - loader for library
 *
 * Copyright (C) 2019 Thorsten Otto
 */
#include "slb/freetype.h"
#include <stdio.h>
#include <stdarg.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <fcntl.h>
#include <time.h>
#include <errno.h>
#include <mint/cookie.h>
#include <mint/mintbind.h>
#include <math.h>
#include <setjmp.h>
#include <freetype/fttypes.h>
#include "libstrct.h"
#include "slbload.h"
#include <slb/zlib.h>
#include <slb/png.h>
#include <slb/bzip2.h>
#include "vernum.h"

struct _freetype_funcs freetype_funcs;


typedef struct
{
	long key;
	long value;
} COOKIE;


static COOKIE *get_cookie_jar(void)
{
	return (COOKIE *) Setexc(360, (void (*)()) -1L);
}


static int get_cookie(long key, long *value)
{
	COOKIE *cookiejar = get_cookie_jar();

	*value = 0;
	if (cookiejar)
	{
		while (cookiejar->key)
		{
			if (cookiejar->key == key)
			{
				*value = cookiejar->value;
				return 1;
			}
			cookiejar++;
		}
	}
	return 0;
}


#if defined(__MSHORT__) || defined(__PUREC__) || defined(__AHCC__)

typedef FT_Int slb_int_t;

static void *__CDECL gcc_memset(void *s, slb_int_t val, size_t len)
{
	return memset(s, (int)val, len);
}

static void *__CDECL gcc_memcpy(void *d, const void *s, size_t len)
{
	return memcpy(d, s, len);
}

static slb_int_t __CDECL gcc_memcmp(const void *s, const void *d, size_t len)
{
	return memcmp(s, d, len);
}

static void *__CDECL gcc_memchr(const void *s, slb_int_t c, size_t n)
{
	return memchr(s, (int)c, n);
}

static size_t __CDECL gcc_strlen(const char *s)
{
	return strlen(s);
}

static char *__CDECL gcc_strcpy(char *d, const char *s)
{
	return strcpy(d, s);
}

static char *__CDECL gcc_strncpy(char *d, const char *s, size_t n)
{
	return strncpy(d, s, n);
}

static char *__CDECL gcc_strcat(char *d, const char *s)
{
	return strcat(d, s);
}

static slb_int_t __CDECL gcc_strcmp(const char *d, const char *s)
{
	return strcmp(d, s);
}

static slb_int_t __CDECL gcc_strncmp(const char *d, const char *s, size_t n)
{
	return strncmp(d, s, n);
}

static char *__CDECL gcc_strstr(const char *d, const char *s)
{
	return strstr(d, s);
}

static char *__CDECL gcc_strrchr(const char *d, slb_int_t c)
{
	return strrchr(d, (int)c);
}

static long __CDECL gcc_strtol(const char *s, char **end, slb_int_t base)
{
	return strtol(s, end, (int)base);
}

static FILE *__CDECL gcc_fopen(const char *fname, const char *mode)
{
	return fopen(fname, mode);
}

static slb_int_t __CDECL gcc_fclose(FILE *fp)
{
	return fclose(fp);
}

static slb_int_t __CDECL gcc_ferror(FILE *fp)
{
	return ferror(fp);
}

static size_t __CDECL gcc_fread(void *buf, size_t size, size_t count, FILE *fp)
{
	return fread(buf, size, count, fp);
}

static size_t __CDECL gcc_fwrite(const void *buf, size_t size, size_t count, FILE *fp)
{
	return fwrite(buf, size, count, fp);
}

static slb_int_t __CDECL gcc_sprintf(char *buf, const char *format, ...)
{
	va_list args;
	int ret;
	va_start(args, fmt);
	ret = vsprintf(buf, format, args);
	va_end(args);
	return ret;
}

static long __CDECL gcc_ftell(FILE *fp)
{
	return ftell(fp);
}

static slb_int_t __CDECL gcc_fseek(FILE *fp, long offset, slb_int_t whence)
{
	return fseek(fp, offset, (int)whence);
}

static void *__CDECL gcc_malloc(size_t s)
{
	return malloc(s);
}

static void __CDECL gcc_free(void *s)
{
	free(s);
}

static void *__CDECL gcc_realloc(void *ptr, size_t s)
{
	return realloc(ptr, s);
}

static slb_int_t __CDECL gcc_sigsetjmp(jmp_buf buf, slb_int_t savesig)
{
	(void)savesig;
	/*
	 * note: mintlib's jmp_buf and Pure-C's have different usages
	 * in the fields, but luckily they are of the same size,
	 * so we can actually do this:
	 */
	return setjmp(buf);
}

static void __CDECL gcc_longjmp(jmp_buf buf, slb_int_t e)
{
	longjmp(buf, (int)e);
}

static char *__CDECL gcc_getenv(const char *name)
{
	return getenv(name);
}

#include "qsort.c"


#if defined(__PUREC__) || defined(__AHCC__)
#define ESRCH	3	    /* No such process            */
#define EINTR	4		/* Interrupted system call    */
#define ENXIO	6		/* No such device or address  */
#define EFAULT	14		/* Bad address                */
#define ENOTBLK	15		/* Bulk device required       */
#define EBUSY	16		/* Resource is busy           */
#define ENOTTY	25		/* Not a terminal             */
#define ETXTBSY	26		/* Text file is busy          */
#define EFBIG	27		/* File is too large          */
#define EMLINK	31		/* Too many links             */
#define EPIPE	32		/* Broken pipe                */
#define EILSEQ  84      /* Illegal byte sequence      */
#endif

#define S(x) freetype_funcs.p_##x = gcc_##x

#else

#define S(x) freetype_funcs.p_##x = x

#endif


static long __CDECL freetype_zlib_open(void)
{
	SLB *freetype = slb_freetype_get();
	SLB *zlib;
	long ret;

	if (!freetype)
		return -EBADF;
	if (!freetype->handle)
		return -EBADF;
	zlib = slb_zlib_get();
	if (!zlib)
		return -EBADF;
	if (zlib->handle == 0)
	{
		ret = slb_zlib_open(freetype_funcs.slbpath);
		if (ret != 0)
			return ret;
		freetype_funcs.zlib_opened_here = 1;
	}
	freetype_funcs.zlib = zlib;
	return 0;
}


static long __CDECL freetype_bzip2_open(void)
{
	SLB *freetype = slb_freetype_get();
	SLB *bzip2;
	long ret;

	if (!freetype)
		return -EBADF;
	if (!freetype->handle)
		return -EBADF;
	bzip2 = slb_bzip2_get();
	if (!bzip2)
		return -EBADF;
	if (bzip2->handle == 0)
	{
		ret = slb_bzip2_open(freetype_funcs.slbpath);
		if (ret != 0)
			return ret;
		freetype_funcs.bzip2_opened_here = 1;
	}
	freetype_funcs.bzip2 = bzip2;
	return 0;
}


static long __CDECL freetype_png_open(void)
{
	SLB *freetype = slb_freetype_get();
	SLB *png;
	long ret;

	if (!freetype)
		return -EBADF;
	if (!freetype->handle)
		return -EBADF;
	png = slb_pnglib_get();
	if (!png)
		return -EBADF;
	if (png->handle == 0)
	{
		ret = slb_pnglib_open(freetype_funcs.slbpath);
		if (ret != 0)
			return ret;
		freetype_funcs.png_opened_here = 1;
	}
	freetype_funcs.png = png;
	return 0;
}


long slb_freetype_open(const char *slbpath)
{
	long ret;
	SLB *freetype = slb_freetype_get();
	
	if (!freetype)
		return -EBADF;
	if (freetype->handle)
		return 0;

	freetype_funcs.struct_size = sizeof(freetype_funcs);
	freetype_funcs.int_size = sizeof(FT_Int);
	freetype_funcs.freetype_vernum = FREETYPE_VERSION;
	freetype_funcs.slbpath = slbpath;

	freetype_funcs.zlib_opened_here = 0;
	freetype_funcs.zlib = 0;
	freetype_funcs.p_slb_freetype_zlib_open = freetype_zlib_open;

	freetype_funcs.bzip2_opened_here = 0;
	freetype_funcs.bzip2 = 0;
	freetype_funcs.p_slb_freetype_bzip2_open = freetype_bzip2_open;

	freetype_funcs.png_opened_here = 0;
	freetype_funcs.png = 0;
	freetype_funcs.p_slb_freetype_png_open = freetype_png_open;

	S(memset);
	S(memcpy);
	S(memcmp);
	S(memchr);

	S(strlen);
	S(strcpy);
	S(strncpy);
	S(strcat);
	S(strcmp);
	S(strncmp);
	S(strstr);
	S(strrchr);
	S(strtol);

	S(malloc);
	S(free);
	S(realloc);

	S(fopen);
	S(fclose);
	S(fread);
	S(fwrite);
	S(sprintf);
	S(ftell);
	S(fseek);

	S(sigsetjmp);
	S(longjmp);
	
	S(qsort);
	S(getenv);
#undef S

	ret = slb_load(FREETYPE_SHAREDLIB_NAME, slbpath, FREETYPE_VERSION, &freetype->handle, &freetype->exec);
	if (ret < 0)
	{
		slb_freetype_close();
		return ret;
	}

	{
		long flags;
		long cpu;

		/*
		 * check compile flags; that function should be as simple as to just return a constant
		 * and we can hopefully call it even on mismatched configurations
		 */
		flags = freetype_slb_control(0, 0);
		get_cookie(C__CPU, &cpu);
		if (cpu >= 20)
		{
			/* should be able to use a 000 library, anyways */
		} else
		{
			if (flags & (1L << 16))
			{
				/* cpu is not 020+, but library was compiled for it */
				slb_freetype_close();
				return -EINVAL;
			}
		}
#if defined(__mcoldfire__)
		/* if cpu is cf, but library was not compiled for it... */
		if (!(flags & (1L << 17)))
#else
		/* if cpu is not cf, but library was compiled for it... */
		if (flags & (1L << 17))
#endif
		{
			slb_freetype_close();
			return -EINVAL;
		}
	}
	
	ret = freetype_slb_control(1, &freetype_funcs);
	if (ret < 0)
	{
		slb_freetype_close();
		return ret;
	}
	
	return ret;
}
